/* Workaround for DCK texture limitation
   By Lee Killough                 11/97

   DCK has a limit of only allowing 768 textures in a wad.

   This program works around it by automatically swapping
   out textures which are not used in the current level.

   Any TEXTURE? lump will be renamed TECTURE, and a new
   TEXTURE lump will be added with only the necessary
   textures. /restore <wad> removes the shortened TEXTURE
   list and restores the previous TEXTURE list.

   Levels can be specified on the command-line, and the
   DCK.STA file is automatically updated with this info.
   If no file or level is specified, it comes from the
   DCK.STA file.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>

struct wad_header 					/* Linked wad files list.*/
{
	char type[4];
	long int num_entries;
	long int dir_start;
};

struct directory 						/* The directory entry header*/
{
	long int start;
	long int length;
	char name[8];
};

struct SideDef
{
   short int xoff;      /* X offset for texture */
   short int yoff;      /* Y offset for texture */
   unsigned char tex1[8];  /* texture name for the part above */
   unsigned char tex2[8];  /* texture name for the part below */
   unsigned char tex3[8];  /* texture name for the regular part */
   short int sector;    /* adjacent sector */
};

#define HASHSIZE 1024

static struct hash {
  struct hash *next;
  unsigned char tex[8];
} *hashtab[HASHSIZE];

static char pathprefix[100];

static char *prefixed(const char *s)
{
 char *p=malloc(strlen(pathprefix)+strlen(s)+1);
 if (!p)
  {
   fprintf(stderr,"Out of Memory\n");
   exit(1);
  }
 strcpy(p,pathprefix);
 return strcat(p,s);
}

static unsigned hashfunc(const unsigned char *tex)
{
 unsigned long hash = 0;
 int i;
 for (i=0;i<8;i++)
   hash = hash*197 + tex[i];
 return hash % HASHSIZE;
}

static int require(const unsigned char *tex)
{
 struct hash **hp = hashtab + hashfunc(tex);
 while (*hp && strncmp((*hp)->tex,tex,8))
   hp=&(*hp)->next;
 if (!*hp)
   if (!(*hp=malloc(sizeof(**hp))))
     {
      fprintf(stderr,"Could not allocate memory for texture names\n");
      return 1;
     }
   else
     {
      (*hp)->next=NULL;
      strncpy((*hp)->tex,tex,8);
     }
 return 0;
}

static int required(const char *tex)
{
 struct hash *hp = hashtab[hashfunc(tex)];
 while (hp && strncmp(hp->tex,tex,8))
   hp=hp->next;
 return hp!=NULL;
}

static int islevel(const char *c)
{
  return
  ((c[0]=='E' || c[0]=='e') && (c[2]=='M' || c[2]=='m') &&
    c[1]>='1' && c[1]<='9' && c[3]>='1' && c[3]<='9' && !c[4] ) ||
  ((c[0]=='M' || c[0]=='m') && (c[1]=='A' || c[1]=='a') &&
   (c[2]=='P' || c[2]=='p') && !c[5] &&
   c[3]>='0' && c[3]<='9' && c[4]>='0' && c[4]<='9');
}

static const char *cmd;

static int strip(const char *wadfilename, const char *level)
{
 FILE *wad = fopen(wadfilename,"r+b"),*res;
 struct wad_header hdr;
 struct directory *direc;
 int i,state;

 if (!wad)
  {
   fprintf(stderr,"Could not open %s\n",wadfilename);
   return 1;
  }

 res = fopen(prefixed("_RESTORE.BAT"),"wt");
 if (!res)
  {
   fprintf(stderr,"Could not create restoration batch file\n");
   fclose(wad);
   return 1;
  }

 if (fread(&hdr,1,sizeof(hdr),wad)!=sizeof(hdr) ||
     (hdr.type[0]!='I' && hdr.type[0]!='P') || hdr.type[1]!='W'
     || hdr.type[2]!='A' || hdr.type[3]!='D')
  {
   fprintf(stderr,"%s does not appear to be a wad file -- bad magic", wadfilename);
   fclose(wad);
   return 1;
  }

 direc = malloc(sizeof(struct directory) * (hdr.num_entries+1));
 if (!direc)
  {
   fprintf(stderr,"Could not allocate memory for wad directory\n");
   fclose(wad);
   return 1;
  }

 fseek(wad,hdr.dir_start,SEEK_SET);
 fread(direc,sizeof(struct directory),hdr.num_entries,wad);

 for (state=i=0;i<hdr.num_entries;i++)
   if (!state)
     state = !strncmp(direc[i].name,level,8);
   else
     if (!strncmp(direc[i].name,"SIDEDEFS",8))
      {
       struct SideDef *sd = malloc(direc[i].length);
       if (!sd)
        {
         fprintf(stderr,"Could not allocate memory for sidedefs\n");
         fclose(wad);
         return 1;
        }
       fseek(wad,direc[i].start,SEEK_SET);
       fread(sd,1,direc[i].length,wad);
       for (i=direc[i].length/sizeof(struct SideDef);i--;)
         if (require(sd[i].tex1) ||
             require(sd[i].tex2) ||
             require(sd[i].tex3) )
          {
           fclose(wad);
           return 1;
          }
       free(sd);
       state=2;
       break;
      }
 if (state<2)
  {
   fprintf(stderr,"Count not find level %s in wad file %s\n",
     level, wadfilename);
   fclose(wad);
   return 1;
  }

 for (i=0;i<hdr.num_entries;i++)
  if (!strncmp(direc[i].name,"TEXTURE",7))
   {
    long ntex,j,used=0,unused=0;
    char *texture;
    if (!(texture=malloc(direc[i].length)))
     {
      fprintf(stderr,"Could not allocate memory for texture lump\n");
      fclose(wad);
      return 1;
     }
    fseek(wad,direc[i].start,SEEK_SET);
    fread(texture,direc[i].length,1,wad);
    memcpy(&ntex,texture,4);
    for (j=0;j<ntex;j++)
     {
      long off;
      memcpy(&off,texture+4+4*j,4);
      if (required(texture+off))
        memcpy(texture+4*++used,texture+4+4*j,4);
      else
        unused++;
     }
    if (unused)
     {
      memcpy(texture,&used,4);
      fseek(wad,0,SEEK_END);
      direc[hdr.num_entries].start=ftell(wad);
      direc[hdr.num_entries].length=direc[i].length;
      memcpy(direc[hdr.num_entries++].name,direc[i].name,8);
      fwrite(texture,1,direc[i].length,wad);
      direc[i].name[2]='C';
      hdr.dir_start=ftell(wad);
      fwrite(direc,sizeof(struct directory),hdr.num_entries,wad);
      fseek(wad,0,SEEK_SET);
      fwrite(&hdr,sizeof(hdr),1,wad);
      fprintf(res,"%s /restore %s\n",cmd,wadfilename);
     }
    free(texture);
    break;
   }
 free(direc);
 fclose(res);
 fclose(wad);
 return 0;
}

static int restore(const char *wadfilename)
{
 FILE *wad = fopen(wadfilename,"r+b");
 struct wad_header hdr;
 struct directory *direc;
 int i,n;

 if (!wad)
  {
   fprintf(stderr,"Could not open %s\n",wadfilename);
   return 1;
  }

 if (fread(&hdr,1,sizeof(hdr),wad)!=sizeof(hdr) ||
     (hdr.type[0]!='I' && hdr.type[0]!='P') || hdr.type[1]!='W'
     || hdr.type[2]!='A' || hdr.type[3]!='D')
  {
   fprintf(stderr,"%s does not appear to be a wad file -- bad magic", wadfilename);
   fclose(wad);
   return 1;
  }

 direc = malloc(sizeof(struct directory) * hdr.num_entries);
 if (!direc)
  {
   fprintf(stderr,"Could not allocate memory for wad directory\n");
   fclose(wad);
   return 1;
  }

 fseek(wad,hdr.dir_start,SEEK_SET);
 fread(direc,sizeof(struct directory),hdr.num_entries,wad);

 for (n=0,i=0;i<hdr.num_entries;i++)
   if (strncmp(direc[i].name,"TEXTURE",7))
    {
     if (!strncmp(direc[i].name,"TECTURE",7))
       direc[i].name[2]='X';
     direc[n++]=direc[i];
    }
 fseek(wad,hdr.dir_start,SEEK_SET);
 fwrite(direc,sizeof(struct directory),hdr.num_entries=n,wad);
 free(direc);
 fseek(wad,0,SEEK_SET);
 fwrite(&hdr,1,sizeof(hdr),wad);
 fclose(wad);
 return 0;
}

static void errexit(void)
{
 FILE *res = fopen(prefixed("_RESTORE.BAT"),"wt");
 if (res)
   fclose(res);
 sleep(2);
 exit(1);
}

int main(int argc, char **argv)
{
 static char buf[5][129], wadfile[129], level[9];
 FILE *fp,*dck;
 int line=0,err=0,dck22,delme=0;
 char *p;

 cmd = p = strdup(*argv);
 for (;*p;p++)
   if (*p=='/') *p='\\';
 if (strrchr(cmd,'\\'))
   strncpy(pathprefix,cmd,strrchr(cmd,'\\')-cmd+1);

 if (argc>1 && !strcmp(argv[1],"/restore"))
  {
   int rc = restore(argv[2]);
   if (rc)
    {
     fprintf(stderr,"Warning: could not restore texture information in %s\n",
       argv[2]);
     sleep(2);
    }
   return rc;
  }

  {
   struct stat s;
   dck22=!stat(prefixed("DCK22.TXT"),&s);
  }

 dck=fopen(prefixed("_DCK.BAT"),"wt");
 if (!dck)
  {
   fputs("Could not create _DCK.BAT file to run DCK.\n",stderr);
   errexit();
  }
 fputs(prefixed("dck32.exe"),dck);
 while (--argc)
   if (**++argv=='/')
    {
     putc(' ',dck);
     fputs(*argv,dck);
    }
   else
    {
     if (dck22)
       delme=1;
     if (islevel(*argv))
       strupr(strcpy(level,*argv));
     else
       strcpy(wadfile,*argv);
   }

 fp = fopen(prefixed("DCK.STA"), "rt");
 if (fp)
  {
   for (line=0; line<5 && fgets(buf[line],sizeof buf[line],fp);line++)
     ;
   fclose(fp);
  }
 if (!(*level   || (!*wadfile && sscanf(buf[4],"%s",level)==1)) ||
     !(*wadfile || sscanf(buf[3],"%s",wadfile)==1))
  {
    fputs("Warning: Could not determine wad file and level to use.\n"
          "Please specify the wad file and level on the command-line.\n"
          "\nFor example:\n\ndckt foo.wad map01\n\n",stderr);
    err=1;
    delme=1;
  }
 if (!dck22)
   fprintf(dck," %s %s\n",wadfile,level);
 fclose(dck);
 if (delme)
  {
   unlink(prefixed("DCK.STA"));
   if (dck22)
    {
     fputs("Note: DCK2.2 has problems accepting wad file arguments on the command-line.\n"
           "Therefore you must also load the correct level manually. If you choose the\n"
           "wrong level, you may get undefined textures.\n",stderr);
     sleep(6);
    }
  }
 if (err || strip(wadfile,level))
   errexit();
 return 0;
}